from django.contrib.auth.backends import ModelBackend

from .handlers import AuthHandler


class AzureBackend(ModelBackend):
    def authenticate(self, request, token=None, *args, **kwargs):
        if not token:  # pragma: no cover
            return
        user = AuthHandler(request).authenticate(token)

        # Return only if `is_active`
        if self.user_can_authenticate(user):
            return user



# flake8: noqa
"""
Django settings for example_project project.

Generated by 'django-admin startproject' using Django 4.0.5.

For more information on this file, see
https://docs.djangoproject.com/en/4.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/4.0/ref/settings/
"""

from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/4.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "codigo_muy_secreto"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "azure_auth",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "azure_auth.middleware.AzureMiddleware",
]

AUTHENTICATION_BACKENDS = ("azure_auth.backends.AzureBackend",)

ROOT_URLCONF = "azure_auth.tests.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "example_project.wsgi.application"


# Database
# https://docs.djangoproject.com/en/4.0/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
    }
}


# Password validation
# https://docs.djangoproject.com/en/4.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/4.0/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/4.0/howto/static-files/

STATIC_URL = "static/"

# Default primary key field type
# https://docs.djangoproject.com/en/4.0/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

AZURE_AUTH = {
    "CLIENT_ID": "dummy_client_id",
    "CLIENT_SECRET": "dummy_client_secret",
    "REDIRECT_URI": "http://mydomain/azure_auth/callback",
    "SCOPES": ["User.Read"],
    "AUTHORITY": "https://login.microsoftonline.com/dummy_tenant_id",
    "LOGOUT_URI": "http://mydomain/logout",
    "PUBLIC_URLS": ["public", "decorator_protected"],
    "PUBLIC_PATHS": ["/public_external"],
    "ROLES": {
        "95170e67-2bbf-4e3e-a4d7-e7e5829fe7a7": "GroupName1",
        "3dc6539e-0589-4663-b782-fef100d839aa": "GroupName2",
    },
    "USERNAME_ATTRIBUTE": "mail",
    "EXTRA_FIELDS": [],
    "USER_MAPPING_FN": "azure_auth.tests.misc.user_mapping_fn",
}
LOGIN_URL = "/azure_auth/login"
LOGIN_REDIRECT_URL = "/"



import functools
from urllib.parse import urlparse

from django.shortcuts import redirect
from django.urls import reverse

from .handlers import AuthHandler


def azure_auth_required(func):
    @functools.wraps(func)
    def _wrapper(request, *args, **kwargs):
        # If the token is valid (or a new valid one can be generated)
        if AuthHandler(request).user_is_authenticated:
            return func(request, *args, **kwargs)
        return redirect(
            f"{reverse('azure_auth:login')}?next={urlparse(request.path).path}"
        )

    return _wrapper



from django.urls import path

from azure_auth.views import azure_auth_callback, azure_auth_login, azure_auth_logout

app_name = "azure_auth"
urlpatterns = [
    path("login", azure_auth_login, name="login"),
    path("logout", azure_auth_logout, name="logout"),
    path("callback", azure_auth_callback, name="callback"),
]



from django.http import HttpResponse
from django.urls import include, path

from azure_auth.decorators import azure_auth_required


def test_view(request):
    return HttpResponse("not a real view")


def public_view(request):
    return HttpResponse("Public view")


def public_external_view(request):
    return HttpResponse("Public external view")


def middleware_protected_view(request):
    return HttpResponse("A view protected by the middleware")


@azure_auth_required
def decorator_protected_view(request):
    return HttpResponse("A view protected by the decorator")


urlpatterns = [
    path("", test_view),
    path("public/", public_view, name="public"),
    path("public_external/", public_external_view, name="public_external"),
    path("decorator_protected/", decorator_protected_view, name="decorator_protected"),
    path(
        "middleware_protected/", middleware_protected_view, name="middleware_protected"
    ),
    path(
        "azure_auth/",
        include("azure_auth.urls"),
    ),
]



import copy
from http import HTTPStatus
from typing import cast
from unittest.mock import patch
from urllib.parse import quote

import msal
import pytest
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser, Group
from django.test import TestCase, TransactionTestCase, override_settings
from django.urls import reverse
from mixer.backend.django import Mixer

from azure_auth.exceptions import TokenError
from azure_auth.handlers import AuthHandler
from azure_auth.utils import EntraStateSerializer

UserModel = get_user_model()


@pytest.mark.django_db
@pytest.mark.usefixtures("auth_flow")
@patch.object(msal, "ConfidentialClientApplication")
class TestLoginView(TestCase):
    # TODO: What about if user is already logged in and hits this view
    # TODO: Any other breaking flows?
    def test_login(self, mocked_msal_app):
        mocked_msal_app.return_value.initiate_auth_code_flow.return_value = (
            self.auth_flow  # type: ignore
        )
        resp = self.client.get(reverse("azure_auth:login"))
        assert resp.status_code == HTTPStatus.FOUND
        assert resp.url == self.auth_flow["auth_uri"]  # type: ignore
        assert self.client.session._session == {"auth_flow": self.auth_flow}  # type: ignore

        # MSAL calls
        mocked_msal_app.assert_called_once_with(
            client_id=settings.AZURE_AUTH["CLIENT_ID"],
            client_credential=settings.AZURE_AUTH["CLIENT_SECRET"],
            authority=settings.AZURE_AUTH["AUTHORITY"],
            # Don't care about the `token_cache` object so just pipe it in
            token_cache=mocked_msal_app.call_args.kwargs["token_cache"],
        )

        mocked_msal_app.return_value.initiate_auth_code_flow.assert_called_once_with(
            scopes=settings.AZURE_AUTH["SCOPES"],
            redirect_uri=settings.AZURE_AUTH["REDIRECT_URI"],
            state='{"next": null}',
            prompt=None,
        )

    def test_login_redirect(self, mocked_msal_app):
        mocked_msal_app.return_value.initiate_auth_code_flow.return_value = (
            self.auth_flow  # type: ignore
        )
        resp = self.client.get(f"{reverse('azure_auth:login')}?next=/dummy_next/")
        assert resp.status_code == HTTPStatus.FOUND
        assert resp.url == self.auth_flow["auth_uri"]  # type: ignore
        assert self.client.session._session == {"auth_flow": self.auth_flow}  # type: ignore

        # MSAL calls
        mocked_msal_app.assert_called_once_with(
            client_id=settings.AZURE_AUTH["CLIENT_ID"],
            client_credential=settings.AZURE_AUTH["CLIENT_SECRET"],
            authority=settings.AZURE_AUTH["AUTHORITY"],
            # Don't care about the `token_cache` object so just pipe it in
            token_cache=mocked_msal_app.call_args.kwargs["token_cache"],
        )

        mocked_msal_app.return_value.initiate_auth_code_flow.assert_called_once_with(
            scopes=settings.AZURE_AUTH["SCOPES"],
            redirect_uri=settings.AZURE_AUTH["REDIRECT_URI"],
            state='{"next": "/dummy_next/"}',
            prompt=None,
        )


@pytest.mark.django_db
@pytest.mark.usefixtures("token")
@pytest.mark.usefixtures("auth_flow")
@patch.object(AuthHandler, "cache")
@patch("azure_auth.handlers.requests")
@patch.object(msal, "ConfidentialClientApplication")
class TestCallbackView(TransactionTestCase):
    def setUp(self):
        super().setUp()
        # Store in variable first
        # https://docs.djangoproject.com/en/4.0/topics/testing/tools/
        session = self.client.session
        session["auth_flow"] = self.auth_flow  # type: ignore
        session.save()

        # Adds a group to existing user that is not part of the token.
        # Should be removed automatically during authentication.
        dummy_group = Group.objects.get_or_create(name="GroupName2")[0]
        self.user.groups.add(dummy_group)  # type: ignore

    def _mocked_response(self, code, json):
        class MockResponse:
            def __init__(self, status_code, data):
                self.status_code = status_code
                self.data = data
                self.ok = True if self.status_code == 200 else False

            def json(self):
                return self.data

        return MockResponse(code, json)

    @staticmethod
    def _get_graph_response(user):
        return {
            "givenName": user.first_name,
            "mail": user.email,
            "surname": user.last_name,
        }

    def _msal_asserts(self, mocked_msal_app):
        # MSAL calls
        mocked_msal_app.assert_called_once_with(
            client_id=settings.AZURE_AUTH["CLIENT_ID"],
            client_credential=settings.AZURE_AUTH["CLIENT_SECRET"],
            authority=settings.AZURE_AUTH["AUTHORITY"],
            # Don't care about the `token_cache` object so just pipe it in
            token_cache=mocked_msal_app.call_args.kwargs["token_cache"],
        )
        m_acf = mocked_msal_app.return_value.acquire_token_by_auth_code_flow
        m_acf.assert_called_once_with(
            auth_code_flow=self.auth_flow,  # type: ignore
            # The request for this view comes from the auth server so just pipe
            # `auth_response` in
            auth_response=m_acf.call_args.kwargs["auth_response"],
        )

    def _graph_asserts(self, mocked_requests):
        # Graph API calls
        mocked_requests.get.assert_called_once_with(
            "https://graph.microsoft.com/v1.0/me",
            headers={"Authorization": f"Bearer {self.token['access_token']}"},  # type: ignore
            params=None,
        )

    def test_callback_existing_user(
        self, mocked_msal_app, mocked_requests, mocked_cache
    ):
        mocked_msal_app.return_value.acquire_token_by_auth_code_flow.return_value = (
            self.token  # type: ignore
        )
        mocked_cache.has_state_changed = True
        mocked_cache.serialize = msal.SerializableTokenCache().serialize
        mocked_cache.deserialize = msal.SerializableTokenCache().deserialize

        # Graph API response
        expected_response_json = self._get_graph_response(self.user)  # type: ignore
        mocked_requests.get.side_effect = [
            self._mocked_response(HTTPStatus.OK, expected_response_json)
        ]

        # Assert user is not staff before log in
        assert self.user.is_staff is False  # type: ignore

        resp = self.client.get(reverse("azure_auth:callback"))
        assert resp.status_code == HTTPStatus.FOUND
        assert resp.url == settings.LOGIN_REDIRECT_URL  # type: ignore
        assert "id_token_claims" in self.client.session

        # Attribute update checks
        self.user.refresh_from_db()  # type: ignore
        assert self.user.is_staff is True  # type: ignore

        # Group creation checks in existing user
        assert Group.objects.filter(name="GroupName1").exists()
        assert Group.objects.filter(name="GroupName2").exists()
        assert self.user.groups.filter(name="GroupName1").exists()  # type: ignore
        assert not self.user.groups.filter(name="GroupName2").exists()  # type: ignore

        self._msal_asserts(mocked_msal_app)
        self._graph_asserts(mocked_requests)

    def test_callback_new_user(self, mocked_msal_app, mocked_requests, mocked_cache):
        mocked_msal_app.return_value.acquire_token_by_auth_code_flow.return_value = (
            self.token  # type: ignore
        )
        mocked_cache.has_state_changed = True
        mocked_cache.serialize = msal.SerializableTokenCache().serialize
        mocked_cache.deserialize = msal.SerializableTokenCache().deserialize

        # Generate unsaved new user instance
        custom_mixer = Mixer(commit=False)
        new_user = cast(AbstractUser, custom_mixer.blend(UserModel))
        assert len(UserModel.objects.all()) == 1

        # Graph API response
        expected_response_json = self._get_graph_response(new_user)
        mocked_requests.get.side_effect = [
            self._mocked_response(HTTPStatus.OK, expected_response_json)
        ]
        resp = self.client.get(reverse("azure_auth:callback"))
        assert resp.status_code == HTTPStatus.FOUND
        assert resp.url == settings.LOGIN_REDIRECT_URL  # type: ignore
        assert "id_token_claims" in self.client.session

        self._msal_asserts(mocked_msal_app)
        self._graph_asserts(mocked_requests)

        # User creation checks
        created_user = cast(
            AbstractUser,
            UserModel._default_manager.get_by_natural_key(new_user.email),  # type: ignore
        )
        assert created_user.username == new_user.email
        assert created_user.first_name == new_user.first_name
        assert created_user.last_name == new_user.last_name
        assert created_user.is_staff is True
        assert created_user.is_active is True

        # Group creation checks
        assert Group.objects.filter(name="GroupName1").exists()
        assert Group.objects.filter(name="GroupName2").exists()
        assert created_user.groups.filter(name="GroupName1").exists()
        assert not created_user.groups.filter(name="GroupName2").exists()

    def test_callback_acquire_token_error(self, mocked_msal_app, *args):
        mocked_msal_app.return_value.acquire_token_by_auth_code_flow.return_value = {
            "error": "dummy_error",
            "error_description": "dummy_error_description",
        }

        with pytest.raises(TokenError) as exc:
            self.client.get(reverse("azure_auth:callback"))
        assert str(exc.value) == "dummy_error\ndummy_error_description"

        self._msal_asserts(mocked_msal_app)

    def test_callback_unauthorized(
        self, mocked_msal_app, mocked_requests, mocked_cache
    ):
        mocked_msal_app.return_value.acquire_token_by_auth_code_flow.return_value = (
            self.token  # type: ignore
        )
        mocked_cache.has_state_changed = True
        mocked_cache.serialize = msal.SerializableTokenCache().serialize
        mocked_cache.deserialize = msal.SerializableTokenCache().deserialize

        # Graph API response
        mocked_requests.get.side_effect = [
            self._mocked_response(
                HTTPStatus.UNAUTHORIZED,
                {"error": {"code": "dummy_error", "message": "dummy_message"}},
            )
        ]

        with pytest.raises(TokenError) as exc:
            self.client.get(reverse("azure_auth:callback"))
        assert str(exc.value) == "dummy_error\ndummy_message"

        self._msal_asserts(mocked_msal_app)
        self._graph_asserts(mocked_requests)

    def test_callback_inactive_user(
        self, mocked_msal_app, mocked_requests, mocked_cache
    ):
        mocked_msal_app.return_value.acquire_token_by_auth_code_flow.return_value = (
            self.token  # type: ignore
        )
        mocked_cache.has_state_changed = True
        mocked_cache.serialize = msal.SerializableTokenCache().serialize
        mocked_cache.deserialize = msal.SerializableTokenCache().deserialize

        # Graph API response
        expected_response_json = self._get_graph_response(self.user)  # type: ignore
        mocked_requests.get.side_effect = [
            self._mocked_response(HTTPStatus.OK, expected_response_json)
        ]

        # Make user inactive
        self.user.is_active = False  # type: ignore
        self.user.save()  # type: ignore
        resp = self.client.get(reverse("azure_auth:callback"))
        assert resp.status_code == HTTPStatus.FORBIDDEN
        assert resp.content.decode() == "Invalid email for this app."

    def test_callback_redirect(self, mocked_msal_app, mocked_requests, mocked_cache):
        mocked_msal_app.return_value.acquire_token_by_auth_code_flow.return_value = (
            self.token  # type: ignore
        )
        mocked_cache.has_state_changed = True
        mocked_cache.serialize = msal.SerializableTokenCache().serialize
        mocked_cache.deserialize = msal.SerializableTokenCache().deserialize

        # Graph API response
        expected_response_json = self._get_graph_response(self.user)  # type: ignore
        mocked_requests.get.side_effect = [
            self._mocked_response(HTTPStatus.OK, expected_response_json)
        ]

        resp = self.client.get(
            f"{reverse('azure_auth:callback')}?state={quote(EntraStateSerializer().serialize(next='/middleware_protected/'), safe='/')}"
        )
        assert resp.status_code == HTTPStatus.FOUND
        assert resp.url == reverse("middleware_protected")  # type: ignore
        assert "id_token_claims" in self.client.session

        self._msal_asserts(mocked_msal_app)
        self._graph_asserts(mocked_requests)


@pytest.mark.usefixtures("token")
class TestLogoutView(TestCase):
    # No redirect logout
    no_logout_query_params_settings = copy.deepcopy(settings.AZURE_AUTH)
    del no_logout_query_params_settings["LOGOUT_URI"]

    def setUp(self):
        super().setUp()
        self.client.force_login(self.user)  # type: ignore

        # Token will be on the session
        session = self.client.session
        session["id_token_claims"] = self.token["id_token_claims"]  # type: ignore
        session.save()

    def test_logout_with_query_params(self):
        # Check user has been correctly logged in
        assert all(
            [
                key in self.client.session
                for key in [
                    "_auth_user_id",
                    "_auth_user_backend",
                    "_auth_user_hash",
                ]
            ]
        )
        resp = self.client.get(reverse("azure_auth:logout"))
        assert resp.status_code == HTTPStatus.FOUND
        assert (
            resp.url == f"{settings.AZURE_AUTH['AUTHORITY']}/oauth2/v2.0/logout"  # type: ignore
            f"?post_logout_redirect_uri=http%3A%2F%2Fmydomain%2Flogout&logout_hint=dummy_id_token_claims_login_hint"
        )
        assert not self.client.session.keys()

    @override_settings(AZURE_AUTH=no_logout_query_params_settings)
    def test_logout_without_query_params(self):
        # Check user has been correctly logged in
        assert all(
            [
                key in self.client.session
                for key in ["_auth_user_id", "_auth_user_backend", "_auth_user_hash"]
            ]
        )
        resp = self.client.get(reverse("azure_auth:logout"))
        assert resp.status_code == HTTPStatus.FOUND
        assert (
            resp.url  # type: ignore
            == f"{settings.AZURE_AUTH['AUTHORITY']}/oauth2/v2.0/logout?logout_hint=dummy_id_token_claims_login_hint"
        )
        assert not self.client.session.keys()



class DjangoAzureAuthException(Exception):
    pass


class TokenError(DjangoAzureAuthException):
    def __init__(self, message, description):
        self.message = message if message else ""
        self.description = description if description else ""

    def __str__(self):
        return f"{self.message}\n{self.description}"



import json


class EntraStateSerializer:
    def serialize(self, **kwargs):
        return json.dumps(kwargs)

    def deserialize(self, state: str):
        try:
            return json.loads(state)
        except json.JSONDecodeError:
            return {}



from typing import Any

import pytest
from django.contrib.auth import get_user_model
from mixer.backend.django import mixer

UserModel = get_user_model()


@pytest.fixture(scope="function", autouse=True)
def user(request):
    _user: Any = mixer.blend(UserModel)
    _user.username = _user.email
    _user.is_staff = False
    _user.save()

    if request.cls:
        request.cls.user = _user
    return _user


@pytest.fixture(scope="function")
def auth_flow(request):
    _auth_flow = {
        "state": "dummy_state",
        "redirect_uri": "http://localhost:8000/azure_auth/callback",
        "scope": ["profile", "offline_access", "User.Read", "openid"],
        "auth_uri": "https://aad-auth-uri.com",
    }

    if request.cls:
        request.cls.auth_flow = _auth_flow
    return _auth_flow


@pytest.fixture(scope="function")
def token(request):
    _token = {
        "token_type": "Bearer",
        "scope": "openid profile User.Read email",
        "expires_in": 3749,
        "ext_expires_in": 3749,
        "access_token": "dummy_access_token",
        "refresh_token": "dummy_refresh_token",
        "id_token": "dummy_id_token",
        "client_info": "dummy_client_info",
        "id_token_claims": {
            "aud": "dummy_id_token_claims_aud",
            "iss": "dummy_id_token_claims_iss",
            "iat": 1655968842,
            "nbf": 1655968842,
            "exp": 1655972742,
            "aio": "dummy_id_token_claims_aio",
            "name": "dummy_id_token_claims_name",
            "nonce": "dummy_id_token_claims_nonce",
            "oid": "dummy_id_token_claims_oid",
            "preferred_username": "dummy_id_token_claims_preferred_username",
            "rh": "dummy_id_token_claims_rh",
            "sub": "dummy_id_token_claims_sub",
            "tid": "dummy_id_token_claims_tid",
            "uti": "dummy_id_token_claims_uti",
            "ver": "2.0",
            "roles": ["95170e67-2bbf-4e3e-a4d7-e7e5829fe7a7"],
            "login_hint": "dummy_id_token_claims_login_hint",
        },
    }

    if request.cls:
        request.cls.token = _token
    return _token



from collections import ChainMap

import pytest
from django.conf import settings
from django.contrib.sessions.middleware import SessionMiddleware
from django.http import HttpResponse
from django.test import RequestFactory, TestCase, override_settings
from django.urls import reverse_lazy

from azure_auth.handlers import AuthHandler


@pytest.mark.django_db
@pytest.mark.usefixtures("token")
class TestAzureAuthHandler(TestCase):
    def setUp(self):
        self.request_factory = RequestFactory()
        self.session_middleware = SessionMiddleware(lambda x: HttpResponse())

    @override_settings(
        AZURE_AUTH=ChainMap(
            {"REDIRECT_URI": "http://testserver/azure_auth/callback_absolut"},
            settings.AZURE_AUTH,
        )
    )
    def test_callback_uri_absolut(self):
        handler = self._build_auth_handler()
        redirect_uri = handler._get_redirect_uri()
        self.assertEqual(redirect_uri, "http://testserver/azure_auth/callback_absolut")

    @override_settings(
        AZURE_AUTH=ChainMap(
            {"REDIRECT_URI": "/azure_auth/callback_relative"}, settings.AZURE_AUTH
        )
    )
    def test_callback_uri_relative(self):
        handler = self._build_auth_handler()
        redirect_uri = handler._get_redirect_uri()
        self.assertEqual(redirect_uri, "http://testserver/azure_auth/callback_relative")

    @override_settings(
        AZURE_AUTH=ChainMap(
            {"REDIRECT_URI": reverse_lazy("decorator_protected")}, settings.AZURE_AUTH
        )
    )
    def test_callback_uri_reverse_lazy(self):
        handler = self._build_auth_handler()
        redirect_uri = handler._get_redirect_uri()
        self.assertEqual(redirect_uri, "http://testserver/decorator_protected/")

    def _build_auth_handler(self) -> AuthHandler:
        req = self.request_factory.get("/")
        self.session_middleware.process_request(req)
        return AuthHandler(req)

    def test_group_mapping(self):
        handler = self._build_auth_handler()
        handler.sync_groups(self.user, self.token)
        self.assertEqual(self.user.groups.count(), 1)

    @override_settings(
        AZURE_AUTH=ChainMap(
            {
                "ROLES": {
                    # Just some random GUID, that is not in the token
                    "a1ad6d75-11c5-442b-9b32-17bdebe82427": "GroupName1",
                }
            },
            settings.AZURE_AUTH,
        )
    )
    def test_no_mapping(self):
        handler = self._build_auth_handler()
        handler.sync_groups(self.user, self.token)
        self.assertEqual(self.user.groups.count(), 0)

    @override_settings(
        AZURE_AUTH=ChainMap(
            {
                "ROLES": {
                    "95170e67-2bbf-4e3e-a4d7-e7e5829fe7a7": ["GroupName1"],
                }
            },
            settings.AZURE_AUTH,
        )
    )
    def test_single_group_list_mapping(self):
        handler = self._build_auth_handler()
        handler.sync_groups(self.user, self.token)
        self.assertEqual(self.user.groups.count(), 1)

    @override_settings(
        AZURE_AUTH=ChainMap(
            {
                "ROLES": {
                    "95170e67-2bbf-4e3e-a4d7-e7e5829fe7a7": [
                        "GroupName1",
                        "GroupName2",
                    ],
                }
            },
            settings.AZURE_AUTH,
        )
    )
    def test_multi_group_list_mapping(self):
        handler = self._build_auth_handler()
        handler.sync_groups(self.user, self.token)
        self.assertEqual(self.user.groups.count(), 2)

    @override_settings(
        AZURE_AUTH=ChainMap(
            {
                "ROLES": {
                    "95170e67-2bbf-4e3e-a4d7-e7e5829fe7a7": [
                        "GroupName1",
                        "",
                    ],
                }
            },
            settings.AZURE_AUTH,
        )
    )
    def test_multi_group_list_mapping_with_empty(self):
        handler = self._build_auth_handler()
        handler.sync_groups(self.user, self.token)
        self.assertEqual(self.user.groups.count(), 1)

    @override_settings(
        AZURE_AUTH=ChainMap(
            {
                "ROLES": {
                    "95170e67-2bbf-4e3e-a4d7-e7e5829fe7a7": [
                        "GroupName1",
                        None,
                    ],
                }
            },
            settings.AZURE_AUTH,
        )
    )
    def test_multi_group_list_mapping_with_none(self):
        handler = self._build_auth_handler()
        handler.sync_groups(self.user, self.token)
        self.assertEqual(self.user.groups.count(), 1)

    @override_settings(
        AZURE_AUTH=ChainMap(
            {
                "ROLES": {
                    "95170e67-2bbf-4e3e-a4d7-e7e5829fe7a7": [],
                }
            },
            settings.AZURE_AUTH,
        )
    )
    def test_empty_group_list_mapping(self):
        handler = self._build_auth_handler()
        handler.sync_groups(self.user, self.token)
        self.assertEqual(self.user.groups.count(), 0)



from django.apps import AppConfig
from django.conf import settings
from django.core.checks import Error, register


@register()
def azure_auth_check(app_configs, **kwargs):
    errors = []

    if not settings.AZURE_AUTH.get("USERNAME_ATTRIBUTE"):
        errors.append(
            Error(
                "misconfigured settings",
                hint="Specify a value for `USERNAME_ATTRIBUTE`.",
                obj=settings.AZURE_AUTH,
                id="azure_auth.E001",
            )
        )
    return errors


class AzureAuthConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "azure_auth"



import datetime
import importlib
from http import HTTPStatus
from typing import Optional, cast
from urllib import parse

import msal
import requests
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractBaseUser, Group
from django.http import HttpRequest

from azure_auth.exceptions import DjangoAzureAuthException, TokenError

UserModel = cast(AbstractBaseUser, get_user_model())


class AuthHandler:
    """
    Class to interface with `msal` package and execute authentication process.
    """

    def __init__(self, request: HttpRequest):
        """

        :param request: HttpRequest
        """
        self.request = request
        self.graph_user_endpoint = settings.AZURE_AUTH.get(
            "GRAPH_USER_ENDPOINT", "https://graph.microsoft.com/v1.0/me"
        )
        self.auth_flow_session_key = "auth_flow"
        self._cache = msal.SerializableTokenCache()
        self._msal_app = None

        # Eagerly load the claims from the session
        self.claims = self.request.session.get("id_token_claims", {})

    def get_auth_uri(self, state: Optional[str] = None) -> str:
        """
        Requests the auth flow dictionary and stores it on the session to be
        queried later in the auth process.

        :param state: State to persist during log in
        :return: Authentication redirect URI
        """
        redirect_uri = self._get_redirect_uri()
        flow = self.msal_app.initiate_auth_code_flow(
            scopes=settings.AZURE_AUTH["SCOPES"],
            redirect_uri=redirect_uri,
            state=state,
            prompt=settings.AZURE_AUTH.get("PROMPT", None),
        )
        self.request.session[self.auth_flow_session_key] = flow
        return flow["auth_uri"]

    def get_token_from_flow(self):
        """
        Acquires the token from the auth flow on the session and the content of
        the redirect request from Active Directory.

        :return: Token result containing `access_token`/`id_token` and other
        claims, depending on scopes used
        """
        flow = self.request.session.pop(self.auth_flow_session_key, {})
        token_result = self.msal_app.acquire_token_by_auth_code_flow(
            auth_code_flow=flow, auth_response=self.request.GET
        )
        if "error" in token_result:
            raise TokenError(token_result["error"], token_result["error_description"])
        self._save_cache()
        self.request.session["id_token_claims"] = token_result["id_token_claims"]
        return token_result

    def get_token_from_cache(self):
        accounts = self.msal_app.get_accounts()
        if accounts:  # pragma: no branch
            # Will return `None` if CCA cannot retrieve or generate new token
            token_result = self.msal_app.acquire_token_silent(
                scopes=settings.AZURE_AUTH["SCOPES"], account=accounts[0]
            )
            self._save_cache()

            # `acquire_token_silent` doesn't always return ID token/ID token claims
            # https://github.com/AzureAD/microsoft-authentication-library-for-python/issues/139
            if token_result and token_result.get("id_token_claims"):
                self.request.session["id_token_claims"] = token_result[
                    "id_token_claims"
                ]
            return token_result

    def authenticate(self, token: dict) -> AbstractBaseUser:
        """
        Helper method to authenticate the user. Gets the Azure user from the
        Microsoft Graph endpoint and gets/creates the associated Django user.

        :param token: MSAL auth token dictionary
        :return: Django user instance
        """
        azure_user = self._get_azure_user(token["access_token"])

        # Get extra fields
        extra_fields = {}
        if fields := settings.AZURE_AUTH.get("EXTRA_FIELDS"):  # pragma: no branch
            extra_fields = self._get_azure_user(token["access_token"], fields=fields)

        # Combine user profile attributes, extra attributes and ID token claims
        # https://learn.microsoft.com/en-us/entra/external-id/customers/concept-user-attributes
        # https://learn.microsoft.com/en-us/entra/identity-platform/id-token-claims-reference
        attributes = {**azure_user, **extra_fields, **token.get("id_token_claims", {})}
        natural_key = attributes[settings.AZURE_AUTH["USERNAME_ATTRIBUTE"]]
        try:
            user = UserModel._default_manager.get_by_natural_key(natural_key)  # type: ignore

            # Sync Django user with AAD attributes
            self._update_user(user, **attributes)
        except UserModel.DoesNotExist:
            user = UserModel._default_manager.create_user(  # type: ignore
                **{UserModel.USERNAME_FIELD: natural_key},  # type: ignore
                **self._map_attributes_to_user(**attributes),
            )

        user = self.sync_groups(user, token)

        return user

    # Syncing azure token claim roles with django user groups
    # A role mapping in the AZURE_AUTH settings is expected.
    # The attribute of the token to use for group membership can be specified
    #   in AZURE_AUTH.GROUP_ATTRIBUTE
    def sync_groups(self, user, token):
        role_mappings = settings.AZURE_AUTH.get("ROLES")
        groups_attr = settings.AZURE_AUTH.get("GROUP_ATTRIBUTE", "roles")
        azure_token_roles = token.get("id_token_claims", {}).get(groups_attr, None)
        if role_mappings:  # pragma: no branch
            for role, group_names in role_mappings.items():
                if not isinstance(group_names, list):
                    group_names = [group_names]
                for group_name in group_names:
                    if not group_name:
                        continue  # Skip empty group names
                    # all groups are created by default if they not exist
                    django_group = Group.objects.get_or_create(name=group_name)[0]

                    if azure_token_roles and role in azure_token_roles:
                        # Add user with permissions to the corresponding django group
                        user.groups.add(django_group)
                    else:
                        # No permission so check if user is in group and remove
                        if user.groups.filter(name=group_name).exists():
                            user.groups.remove(django_group)

        return user

    def get_logout_uri(self) -> str:
        """
        Forms the URI to log the user out in the Active Directory app and
        redirect to the webapp logout page.

        :return: Active Directory app logout URI
        """
        authority = settings.AZURE_AUTH["AUTHORITY"]
        _query_params = {
            "post_logout_redirect_uri": settings.AZURE_AUTH.get("LOGOUT_URI"),
            "logout_hint": self.claims.get("login_hint"),
        }
        query_params = {k: v for k, v in _query_params.items() if v}
        return f"{authority}/oauth2/v2.0/logout?{parse.urlencode(query_params)}"

    @property
    def user_is_authenticated(self) -> bool:
        now = datetime.datetime.now(datetime.timezone.utc).timestamp()

        # Check the ID token is still valid in the first instance
        if now < self.claims.get("exp", 0) and self.request.user.is_authenticated:
            return True

        # Otherwise try refresh the token
        return (
            self.get_token_from_cache() is not None
            and self.request.user.is_authenticated
        )

    def _get_confidential_client(self):
        secret = settings.AZURE_AUTH.get("CLIENT_SECRET", "<client_secret>")
        if secret == "<client_secret>":
            raise DjangoAzureAuthException(
                "CLIENT_TYPE='confidential_client' also requires CLIENT_SECRET to be set in AZURE_AUTH"
            )
        additional_kwargs = settings.AZURE_AUTH.get("ADDITIONAL_CLIENT_KWARGS", {})
        return msal.ConfidentialClientApplication(
            client_id=settings.AZURE_AUTH["CLIENT_ID"],
            client_credential=settings.AZURE_AUTH["CLIENT_SECRET"],
            authority=settings.AZURE_AUTH["AUTHORITY"],
            token_cache=self.cache,
            **additional_kwargs,
        )

    def _get_public_client(self):
        additional_kwargs = settings.AZURE_AUTH.get("ADDITIONAL_CLIENT_KWARGS", {})
        return msal.PublicClientApplication(
            client_id=settings.AZURE_AUTH["CLIENT_ID"],
            authority=settings.AZURE_AUTH["AUTHORITY"],
            token_cache=self.cache,
            **additional_kwargs,
        )

    @property
    def msal_app(self):
        if self._msal_app is None:
            client_type = settings.AZURE_AUTH.get("CLIENT_TYPE", "confidential_client")
            if client_type == "confidential_client":
                self._msal_app = self._get_confidential_client()
            elif client_type == "public_client":
                self._msal_app = self._get_public_client()
            else:
                raise DjangoAzureAuthException(
                    f"Invalid CLIENT_TYPE '{client_type}' specified in AZURE_AUTH settings."
                )
        return self._msal_app

    @property
    def cache(self):
        if self.request.session.get("token_cache"):  # pragma: no branch
            self._cache.deserialize(self.request.session["token_cache"])
        return self._cache

    def _save_cache(self):
        if self.cache.has_state_changed:
            self.request.session["token_cache"] = self.cache.serialize()

    def _get_azure_user(self, token: str, fields: Optional[dict] = None):
        params = {"$select": ",".join(fields)} if fields else None
        resp = requests.get(
            self.graph_user_endpoint,
            headers={"Authorization": f"Bearer {token}"},
            params=params,
        )
        if resp.ok:
            return resp.json()
        elif resp.status_code == HTTPStatus.UNAUTHORIZED:
            error = resp.json()["error"]
            raise TokenError(message=error["code"], description=error["message"])
        else:  # pragma: no cover
            raise DjangoAzureAuthException("An unknown error occurred.")

    def _map_attributes_to_user(self, **fields) -> dict:
        if user_mapping_fn := settings.AZURE_AUTH.get(
            "USER_MAPPING_FN"
        ):  # pragma: no branch
            path, fn = user_mapping_fn.rsplit(".", 1)
            mod = importlib.import_module(path)
            return getattr(mod, fn)(**fields)
        return {}  # pragma: no cover

    def _update_user(self, user: AbstractBaseUser, **fields):
        if user_mapping_fn := settings.AZURE_AUTH.get(  # pragma: no branch
            "USER_MAPPING_FN"
        ):
            path, fn = user_mapping_fn.rsplit(".", 1)
            mod = importlib.import_module(path)
            for field, value in getattr(mod, fn)(**fields).items():
                setattr(user, field, value)
            user.save()

    def _get_redirect_uri(self) -> str:
        redirect_uri = settings.AZURE_AUTH["REDIRECT_URI"]
        if not isinstance(redirect_uri, str):
            # Resolve the URI when it's a reverse_lazy callable
            redirect_uri = str(redirect_uri)
        if not redirect_uri.startswith("http"):
            redirect_uri = self.request.build_absolute_uri(redirect_uri)
        return redirect_uri



from http import HTTPStatus
from unittest.mock import patch

import msal
import pytest
from django.test import TransactionTestCase
from django.urls import reverse


@pytest.mark.django_db
@pytest.mark.usefixtures("token")
@patch.object(msal, "ConfidentialClientApplication")
class TestAzureAuthDecorator(TransactionTestCase):
    def test_invalid_token(self, mocked_msal_app):
        mocked_msal_app.return_value.acquire_token_silent.return_value = None
        resp = self.client.get(reverse("decorator_protected"))
        assert resp.status_code == HTTPStatus.FOUND
        assert resp.url == f"{reverse('azure_auth:login')}?next=/decorator_protected/"  # type: ignore

    def test_valid_token_unauthenticated_user(self, mocked_msal_app):
        # Not sure how this situation could arise but test anyway...

        mocked_msal_app.return_value.acquire_token_silent.return_value = self.token  # type: ignore
        resp = self.client.get(reverse("decorator_protected"))
        assert resp.status_code == HTTPStatus.FOUND
        assert resp.url == f"{reverse('azure_auth:login')}?next=/decorator_protected/"  # type: ignore

    def test_valid_token_authenticated_user(self, mocked_msal_app):
        mocked_msal_app.return_value.acquire_token_silent.return_value = self.token  # type: ignore
        self.client.force_login(self.user)  # type: ignore

        resp = self.client.get(reverse("decorator_protected"))
        assert resp.status_code == HTTPStatus.OK



from urllib.parse import unquote

from django.conf import settings
from django.contrib.auth import authenticate, login, logout
from django.http import HttpRequest, HttpResponseForbidden, HttpResponseRedirect
from django.utils.http import url_has_allowed_host_and_scheme

from azure_auth.utils import EntraStateSerializer

from .handlers import AuthHandler

serializer = EntraStateSerializer()


def azure_auth_login(request: HttpRequest):
    return HttpResponseRedirect(
        AuthHandler(request).get_auth_uri(
            state=serializer.serialize(next=request.GET.get("next"))
        )
    )


def azure_auth_logout(request: HttpRequest):
    # Auth handler has to be initialized before `logout()` to load the claims from the session
    auth_handler = AuthHandler(request)

    logout(request)
    return HttpResponseRedirect(auth_handler.get_logout_uri())


def azure_auth_callback(request: HttpRequest):
    token = AuthHandler(request).get_token_from_flow()
    user = authenticate(request, token=token)
    if user:
        login(request, user)

        # Get `state` query param returned by AAD
        next = serializer.deserialize(unquote(request.GET.get("state", ""))).get(
            "next", ""
        )
        if url_has_allowed_host_and_scheme(next, allowed_hosts=None):
            return HttpResponseRedirect(next)
        return HttpResponseRedirect(settings.LOGIN_REDIRECT_URL)
    return HttpResponseForbidden("Invalid email for this app.")



from urllib.parse import urlparse

from django.conf import settings
from django.http import HttpRequest
from django.shortcuts import redirect
from django.urls import reverse

from .handlers import AuthHandler


class AzureMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        public_views = ["azure_auth:login", "azure_auth:logout", "azure_auth:callback"]
        public_views.extend(settings.AZURE_AUTH.get("PUBLIC_URLS", []))
        self.public_urls = [reverse(view_name) for view_name in public_views]
        self.public_paths = settings.AZURE_AUTH.get(
            "PUBLIC_PATHS", []
        )  # added to resolve paths

    def __call__(self, request: HttpRequest):
        if request.path_info in self.public_urls:
            return self.get_response(request)

        # Added to resolve paths that can't be reversed
        for path in self.public_paths:
            if request.path_info.startswith(path):
                return self.get_response(request)

        if AuthHandler(request).user_is_authenticated:
            return self.get_response(request)

        return redirect(
            f"{reverse('azure_auth:login')}?next={urlparse(request.path).path}"
        )



import datetime
from http import HTTPStatus
from unittest.mock import patch

import msal
import pytest
from django.test import TransactionTestCase
from django.urls import reverse


@pytest.mark.django_db
@pytest.mark.usefixtures("token")
@patch.object(msal, "ConfidentialClientApplication")
class TestAzureAuthMiddleware(TransactionTestCase):
    def test_invalid_token(self, mocked_msal_app):
        mocked_msal_app.return_value.acquire_token_silent.return_value = None
        resp = self.client.get(reverse("middleware_protected"))
        assert resp.status_code == HTTPStatus.FOUND
        assert resp.url == f"{reverse('azure_auth:login')}?next=/middleware_protected/"  # type: ignore

    def test_valid_token_unauthenticated_user(self, mocked_msal_app):
        # Not sure how this situation could arise but test anyway...

        mocked_msal_app.return_value.acquire_token_silent.return_value = self.token  # type: ignore
        resp = self.client.get(reverse("middleware_protected"))
        assert resp.status_code == HTTPStatus.FOUND
        assert resp.url == f"{reverse('azure_auth:login')}?next=/middleware_protected/"  # type: ignore

    def test_valid_token_authenticated_user(self, mocked_msal_app):
        mocked_msal_app.return_value.acquire_token_silent.return_value = self.token  # type: ignore
        self.client.force_login(self.user)  # type: ignore

        resp = self.client.get(reverse("middleware_protected"))
        assert resp.status_code == HTTPStatus.OK
        assert mocked_msal_app.return_value.acquire_token_silent.call_count == 1

    def test_valid_id_token_claims_authenticated_user(self, mocked_msal_app):
        self.client.force_login(self.user)  # type: ignore

        # Set up valid id token claims in session
        session = self.client.session
        session["id_token_claims"] = {
            "exp": datetime.datetime.now(datetime.timezone.utc).timestamp() + 10
        }
        session.save()

        resp = self.client.get(reverse("middleware_protected"))
        assert resp.status_code == HTTPStatus.OK
        assert mocked_msal_app.return_value.acquire_token_silent.call_count == 0

    def test_expired_id_token_claims_authenticated_user(self, mocked_msal_app):
        # Set up mocked incoming token
        new_token_expiry = (
            datetime.datetime.now(datetime.timezone.utc).timestamp() + 1000
        )
        new_token = self.token  # type: ignore
        new_token["id_token_claims"]["exp"] = new_token_expiry

        mocked_msal_app.return_value.acquire_token_silent.return_value = new_token
        self.client.force_login(self.user)  # type: ignore

        # Set up expired id token claims in session
        session = self.client.session
        session["id_token_claims"] = {
            "exp": datetime.datetime.now(datetime.timezone.utc).timestamp() - 10
        }
        session.save()

        resp = self.client.get(reverse("middleware_protected"))
        assert resp.status_code == HTTPStatus.OK
        assert mocked_msal_app.return_value.acquire_token_silent.call_count == 1
        assert self.client.session["id_token_claims"]["exp"] == new_token_expiry

    def test_public_view(self, mocked_msal_app):
        mocked_msal_app.return_value.acquire_token_silent.return_value = None
        resp = self.client.get(reverse("public"))
        assert resp.status_code == HTTPStatus.OK

    def test_public_external_view(self, mocked_msal_app):
        mocked_msal_app.return_value.acquire_token_silent.return_value = None
        resp = self.client.get(reverse("public_external"))
        assert resp.status_code == HTTPStatus.OK
